/******************************************************************************
*
*  decode.cmd
*
*  Decode MP1/2/3 files and store ID3 V1.x Tag information in EA's
*  Decode Ogg-Vorbis files and store vorbis comments in EA's
*
*  Update:  25.09.2002
*  (C)      Marcel Mller 2002
*
******************************************************************************/

IF RxFuncAdd("SysLoadFuncs", "RexxUtil", "SysLoadFuncs") = 0 THEN
   CALL SysLoadFuncs
IF RxFuncAdd("MMLoadFuncs", "RxMMutl", "MMLoadFuncs") = 0 THEN
   CALL MMLoadFuncs

NUMERIC DIGITS 12


IF ARG(1) = '' THEN DO
   SAY "usage: mp3decode [options] file[s]"
   SAY "options:"
   SAY " /Ofile   Output file. If file is a directory the original file name(s) are"
   SAY "          preserved, except for the file extension. By default the files are"
   SAY "          created in the same directory as the original files. Use /O. to"
   SAY "          force the current working directory instead."
   EXIT 48
   END


/* some defaults */
opt.output = ''
opt.trunc = 0
nfiles = 0

/****************************************************************************
   parse command line
*/
params = STRIP(ARG(1))
DO WHILE params \= ''
   IF LEFT(params,1) = '"' THEN DO
      PARSE VAR params '"'param'"'params
      params = STRIP(params, 'L')
      END
    ELSE
      PARSE VAR params param params
   /* param is the next parameter to process ... */
   IF (LEFT(param, 1) = '/') | (LEFT(param, 1) = '-') THEN DO
      /* parsing options ... */
      opt = TRANSLATE(SUBSTR(param, 2, 1))
      SELECT
       WHEN opt = 'O' THEN
         opt.output = SUBSTR(param, 3)
       OTHERWISE
         CALL Error 44, "illegal option '"param"'"
         END
      ITERATE
      END
   /* any other arg ... */
   SAY "Decoding "param"..."

   /* fetch meta information */
   fform = ReadTag(param)

   /* output file name */
   ofile = param
   IF LASTPOS('.',ofile) = LENGTH(ofile)-3 THEN
   ofile = DELSTR(ofile, LENGTH(ofile)-3)
   ofile = ofile'.WAV'
   /* folder */
   IF opt.output \= '' THEN
      ofile = opt.output'\'FILESPEC('N', ofile)

   /* decode ! */
   SELECT
    WHEN fform = 'MPEG' | fform = 'MPEG-ID3' THEN DO
      /* Work around for lame --decode bug when ID3 tags contain not sync-safe binary data */
      IF fform = 'MPEG-ID3' THEN DO
         CALL STREAM param, 'c', 'open read'
         tmp = CHARIN(param, 1, 10)
         id3size = 2097152*C2D(BITAND(SUBSTR(tmp, 7, 1),'7F'x)) + 16384*C2D(BITAND(SUBSTR(tmp, 8, 1),'7F'x)) + 128*C2D(BITAND(SUBSTR(tmp, 9, 1),'7F'x)) + C2D(BITAND(SUBSTR(tmp, 10, 1),'7F'x)) + 10
         CALL STREAM param, 'c', 'close'
         CALL SysSetPriority 1, 3
         '@subfile 'Escape(param) id3size' 0x7ffffff0 | call lame --decode --mp3input - 'Escape(ofile)
         CALL SysSetPriority 2, 0
         END
      ELSE DO
         CALL SysSetPriority 1, 3
         '@call lame --decode --mp3input 'Escape(param) Escape(ofile)
         CALL SysSetPriority 2, 0
         END
      IF RC \= 0 THEN DO
         CALL Warning "lame returned a bad status: "RC
         ITERATE
         END
      /* work around for lame bug: truncate to a multiple of the MP3 frame length */
      /*hdr = CHARIN(ofile, , 44)
      IF hdr = '' THEN DO
         CALL STREAM ofile, 'c', 'close'
         CALL Warning "Failed to open decoded file "ofile"."
         ITERATE
         END
      IF SUBSTR(hdr, 9, 8) \= "WAVEfmt " | SUBSTR(hdr, 37, 4) \= "data" THEN DO
         CALL STREAM ofile, 'c', 'close'
         CALL Warning "Outputfile "ofile" seems not to be of type wave audio."
         ITERATE
         END
      filelen = C2D(REVERSE(SUBSTR(hdr, 5, 4)))
      datalen = C2D(REVERSE(SUBSTR(hdr, 41, 4)))
      bps = C2D(REVERSE(SUBSTR(hdr, 33, 2)))
      rem = datalen // (1152 * bps);
      IF rem \= 0 THEN DO
         SAY "Lame work around: truncating additional "rem%bps" samples"
         CALL CHAROUT ofile, REVERSE(D2C(datalen-rem, 4)), 41
         CALL CHAROUT ofile, REVERSE(D2C(filelen-rem, 4)), 5
         CALL STREAM ofile, 'c', 'close'
         CALL MMSetFileSize ofile, filelen-rem +8
         END
       ELSE
         CALL STREAM ofile, 'c', 'close'*/
      END

    WHEN fform = 'OGG' THEN DO
      CALL SysSetPriority 1, 3
      '@call oggdec -o 'Escape(ofile) Escape(param)
      CALL SysSetPriority 2, 0
      IF RC \= 0 THEN DO
         CALL Warning "oggdec returned a bad status: "RC
         ITERATE
         END
      END

   OTHERWISE
   ITERATE
   END

   /* store meta information */
   SAY "Copy meta information to EAs"
   CALL StoreEA ofile, 'CDT-Title', 'title'
   CALL StoreEA ofile, 'CDT-Artist', 'artist'
   CALL StoreEA ofile, 'CDT-Album', 'album'
   CALL StoreEA ofile, 'CDT-Year', 'year'
   CALL StoreEA ofile, 'CDT-Track', 'track'
   CALL StoreEA ofile, 'CDT-Comment', 'comment'

   /* done */
   nfiles = nfiles +1
   END

IF nfiles = 0 THEN
   CALL Error 10, "Nothing to do"
SAY nfiles" file(s) decoded successfully"

EXIT


/* tagging functions */
GetFormat: PROCEDURE
/* examine data format
   parameter:
      ARG(1) stream
   return:
      'MPEG' MPEG audio
      'OGG'  ogg vorbis
      ''     error
*/
   tmp = CHARIN(ARG(1), 1, 3)
   SELECT
    WHEN tmp = 'ID3' THEN
      RETURN 'MPEG-ID3'
    WHEN BITAND(tmp, 'FFF000'x) = 'FFF000'x THEN
      RETURN 'MPEG'
    WHEN tmp = 'Ogg' THEN
      RETURN 'OGG'
    OTHERWISE
      SAY "file "ARG(1)" seems not to have a recognized format. ("C2X(tmp)")"
      RETURN ''
      END

ReadTag: PROCEDURE EXPOSE opt. tag.
/* read file information
   parameter:
      ARG(1) stream
   return:
      format OK
      ''     error
   global:
      tag.*  file information
*/
   /* read header, read tag*/
   CALL STREAM ARG(1), 'c', 'open read'
   fform = GetFormat(ARG(1))
   IF fform = '' THEN DO
      CALL STREAM ARG(1), 'c', 'close'
      RETURN ''
      END
   /* read tag */
   DROP tag.
   SELECT
    WHEN fform = 'MPEG' | fform = 'MPEG-ID3' THEN DO
      /* read ID3 V1.x Tag */
      num = STREAM(ARG(1), 'c', 'QUERY SIZE')
      IF num < 1024 THEN DO
         SAY 'Cannot read tag of file 'ARG(1)'.'
         CALL STREAM ARG(1), 'c', 'close'
         RETURN ''
         END
      IF opt.trunc THEN DO
         tag = CHARIN(ARG(1), num-1023, 1024)
         CALL STREAM ARG(1), 'c', 'close'
         p = VERIFY(REVERSE(tag)' ', '00'x)
         IF p > 1024-128 THEN DO
            SAY 'Cannot find tag of file 'ARG(1)'.'
            RETURN ''
            END
         p = POS('TAG', tag, 1025-128 - p)
         IF p = 0 THEN DO
            SAY 'Cannot find tag of file 'ARG(1)'.'
            RETURN ''
            END
         CALL MMSetFileSize ARG(1), num-1025+p+128
         IF result \= 0 THEN
            SAY 'Failed to truncate 'ARG(1)', error 'result'.'
         tag = SUBSTR(tag, p, 128)
         END
       ELSE DO
         tag = CHARIN(ARG(1), num-127, 128)
         CALL STREAM ARG(1), 'c', 'close'
         IF SUBSTR(tag, 1, 3) \= 'TAG' THEN DO
            SAY 'File 'ARG(1)' has no tag information.'
            RETURN fform
            END
         END
      /* parse tag, CP translation */
      CALL SetTag 'title', SUBSTR(tag, 4, 30)
      CALL SetTag 'artist', SUBSTR(tag, 34, 30)
      CALL SetTag 'album', SUBSTR(tag, 64, 30)
      CALL SetTag 'year', SUBSTR(tag, 94, 4)
      CALL SetTag 'comment', SUBSTR(tag, 98, 30)
      tmp = RIGHT(C2D(SUBSTR(tag, 127, 1)), 2, '0')
      IF tmp \= 0 & SUBSTR(tag, 126, 1) = '00'x THEN
         tag.track = tmp
      tmp = C2D(RIGHT(tag, 1))
      IF tmp \= 255 THEN
         tag.tmp = tmp
      /* done */
      tag.old = 1
      RETURN fform
      END
    WHEN fform = 'OGG' THEN DO
      CALL STREAM ARG(1), 'c', 'close'
      /* launch vorbiscomment */
      tmpfile = SysTempFileName(VALUE('TEMP', ,'OS2ENVIRONMENT')'\mp3t????.tmp')
      '@vorbiscomment 'Escape(ARG(1))' >'tmpfile
      IF RC \= 0 THEN DO
         SAY 'Execution of vorbiscomment returned a bad status ('RC')'
         pause
         RETURN ''
         END
      /* read result */
      IF STREAM(tmpfile, 'c', 'open read') \= 'READY:' THEN DO
         SAY 'Failed to read result from vorbiscomment ('tmpfile')'
         RETURN ''
         END
      r = 0
      DO FOREVER
         tmp = LINEIN(tmpfile)
         IF tmp = '' THEN LEAVE
         r = 1
         PARSE VAR tmp key'='val
         key = STRIP2(key)
         SELECT
          WHEN key = 'title' | key = 'artist' | key = 'album' | key = 'comment' THEN
            CALL VALUE 'tag.'key, val
          WHEN key = 'tracknumber' THEN
            tag.track = RIGHT(val, 2, '0')
          WHEN key = 'date' THEN
            tag.year = val
          OTHERWISE
            END
         END
      CALL STREAM tmpfile, 'c', 'close'
      CALL SysFileDelete(tmpfile)
      /* done */
      IF r THEN
         RETURN fform
      SAY 'No comments found in file 'ARG(1)'.'
      END
    OTHERWISE NOP
      END
   RETURN ''

SetTag: PROCEDURE EXPOSE tag.
   /* ARG(1) = tag entry
      ARG(2) = value
   */
   p = POS('0'x, ARG(2)'0'x)
   val = STRIP(LEFT(ARG(2), p-1))
   IF val = '' THEN RETURN
   CALL VALUE 'tag.'ARG(1), val
   RETURN

/* quote % characters */
Escape: PROCEDURE
   tmp = ARG(1)
   p = POS('%', tmp)
   DO WHILE p \= 0
      tmp = INSERT('%', tmp, p)
      p = POS('%', tmp, p+2)
      END
   RETURN '"'tmp'"'

IsDir: PROCEDURE
   tmp = STRIP(TRANSLATE(ARG(1), '\', '/'), 'T', '\')
   RETURN FILESPEC('d',ARG(1)) = ARG(1) | (STREAM(tmp, 'c', 'query size') = 0 & STREAM(tmp, 'c', 'query exists') = '')

StoreEA: PROCEDURE EXPOSE tag.
/* write extended attribute
   ARG(1)   file
   ARG(2)   EA name
   ARG(3)   tag name; If the tag.tagname does not exist, the EA is not stored.
*/
   /*SAY "EA: "ARG(3)'_'SYMBOL('tag.'ARG(3))*/
   IF SYMBOL('tag.'ARG(3)) \= 'VAR' THEN RETURN
   val = VALUE('tag.'ARG(3))
   rc = SysPutEA(ARG(1), ARG(2), 'DFFF00000100FDFF'x || REVERSE(D2C(LENGTH(val),2)) || val)
   IF rc \= 0 THEN
      SAY 'failed to store EA information: 'ARG(1)','ARG(2)','val
   RETURN

Strip2: PROCEDURE
   tmp = ARG(1)
   opt = TRANSLATE(LEFT(ARG(2), 1, 'B'))
   IF ARG(3, 'e') THEN
      chr = ARG(2)
    ELSE
      chr = '2009'x
   /* strip front */
   IF TRANSLATE(opt, 'L', 'B') = 'L' THEN DO
      p = VERIFY(tmp, chr)
      IF p = 0 THEN
         RETURN ''
       ELSE
         tmp = SUBSTR(tmp, p)
      END
   /* strip back */
   IF TRANSLATE(opt, 'T', 'B') = 'T' THEN DO
      p = VERIFY(REVERSE(tmp), chr)
      RETURN LEFT(tmp, LENGTH(tmp) - p +1)
      END
   RETURN tmp

/* CP translation ISO to system CP */
Cp2DOS:
   IF \opt.cp THEN RETURN ARG(1)
   RETURN MMTranslateCp(ARG(1), 1004, , '')


/****************************************************************************
   Fehlerbehandlung
*/
Warning: PROCEDURE
   IF ARG(1) \= '' THEN
      CALL LINEOUT STDERR, ARG(1)
   RETURN ARG(1)

Error: PROCEDURE
   CALL LINEOUT STDERR, ARG(2)
   EXIT ARG(1)

